import QtQuick 2.15

Item {
    id: root

    //外部属性
    property string frontImagePath: ""
    property string backImagePath: ""
    property string backgroundOption: "scaled"
    property bool usePureColor: false
    property string bgColor: "black"
    property rect screenGeometry: Qt.rect(0, 0, 0, 0)

    //内部属性
    property real animationDuration: 1000
    property rect sourceRect: Qt.rect(0, 0, 0, 0)

    ParallelAnimation {
        id: animation
        NumberAnimation {
            target: backImage
            property: "opacity"
            from: 1
            to: 0
            duration: animationDuration
            easing.type: Easing.InOutQuad
        }

        NumberAnimation {
            target: frontImage
            property: "opacity"
            from: 0
            to: 1
            duration: animationDuration
            easing.type: Easing.InOutQuad
        }
    }

    Image {
        id: backImage
        anchors.fill: parent
        source: backImagePath ? "file://" + backImagePath : ""
        fillMode: getFillMode(backgroundOption)
        opacity: 0
        visible: !root.usePureColor
    }

    Image {
        id: frontImage
        anchors.fill: parent
        source: frontImagePath ? "file://" + frontImagePath : ""
        fillMode: getFillMode(backgroundOption)
        opacity: 1
        visible: !root.usePureColor
    }

    Rectangle {
        id: colorBackground
        anchors.fill: parent
        color: root.bgColor
        visible: root.usePureColor
    }

    //获取填充模式
    function getFillMode(option) {
        switch(option.toLowerCase()) {
            case "stretched": return Image.Stretch;
            case "scaled": return Image.PreserveAspectCrop;
            case "centered": return Image.Pad;
            case "wallpaper": return Image.Tile;
            case "zoom": return Image.PreserveAspectFit;
            case "spanned": return handleSpanned();
            default: return Image.PreserveAspectCrop;
        }
    }

    //处理跨区
    function handleSpanned() {
        if (backgroundOption != "spanned")
            return Image.Stretch

        var rect = backgroundManager.calculateSpannedSourceRect(screenGeometry, frontImagePath)
        backImage.sourceClipRect = rect
        frontImage.sourceClipRect = rect
        console.log("handleSpanned rect", rect)
        return Image.Stretch
    }

    //更新对齐
    function updateImageAlignment() {
        if (backgroundOption == "wallpaper") {
            backImage.horizontalAlignment = Image.AlignLeft
            backImage.verticalAlignment = Image.AlignTop
            frontImage.horizontalAlignment = Image.AlignLeft
            frontImage.verticalAlignment = Image.AlignTop
        } else {
            backImage.horizontalAlignment = Image.AlignHCenter
            backImage.verticalAlignment = Image.AlignVCenter
            frontImage.horizontalAlignment = Image.AlignHCenter
            frontImage.verticalAlignment = Image.AlignVCenter
        }
    }

    //更换纯色显示
    function updateBackground() {
        if (usePureColor) {
            backImage.visible = false
            frontImage.visible = false
            colorBackground.visible = true
        } else {
            backImage.visible = true
            frontImage.visible = true
            colorBackground.visible = false
        }
    }

    //清除切换跨区后sourceClipRect属性
    function updateImageSourceClipRect() {
        if (backgroundOption != "spanned") {
            backImage.sourceClipRect = undefined
            frontImage.sourceClipRect = undefined
        }
    }

    onFrontImagePathChanged: {
        console.log("onFrontImagePathChanged", backImage.status, frontImage.status)
        if (backImage.status == Image.Ready && frontImage.status == Image.Ready) {
            animation.restart()
        }
        updateImageAlignment()
    }

    onScreenGeometryChanged: {
        handleSpanned()
    }

    Component.onCompleted: {
        updateBackground()
    }

    onUsePureColorChanged: {
        updateBackground()
    }

    onBackgroundOptionChanged: {
        console.log("onBackgroundOptionChanged", backgroundOption)
        frontImage.fillMode = getFillMode(backgroundOption)
        updateImageAlignment()
        updateImageSourceClipRect()
    }

}
