import QtQuick 2.15
import QtQuick.Controls 2.15
import QtGraphicalEffects 1.15

Item {
    id: root
    width: parent.width
    height: editing ? editBackground.height : textItem.height

    // 公共属性
    property var selectedUris: []
    property string display: ""  // 显示的文本内容
    property bool editing: false  // 是否处于编辑状态
    property bool selected: false  // 是否处于选中状态
    property bool isDir: false // 是否处于目录
    property bool isSoftLink: false // 是否是软连接
    property bool showFullText: false

    signal editFinish(var selectedUris, string newText)

    Label {
        id: textItem
        text: root.display
        color: "white"
        anchors.horizontalCenter: parent.horizontalCenter
        anchors.top: parent.top
        anchors.topMargin: 5
        width: parent.width - 8
        wrapMode: Text.WrapAnywhere
        elide: Text.ElideRight
        maximumLineCount: selected && showFullText? 0 : 2
        horizontalAlignment: Text.AlignHCenter
        verticalAlignment: Text.AlignVCenter
        // 只在非编辑状态显示
        visible: !root.editing
    }

    DropShadow {
        anchors.fill: textItem
        source: textItem
        horizontalOffset: 3
        verticalOffset: 3
        layer.enabled: true
        radius: 8
        samples: 16
        color: "#80000000"
        visible: textItem.visible
    }

    // 编辑文本框（编辑状态）
    Rectangle {
        id: editBackground
        x: textEdit.x
        y: textEdit.y
        z: -1
        width: textEdit.width
        height: textEdit.height
        color: "#f0f0f0"  // 背景颜色
        border.color: "gray"  // 边框颜色
        radius: 5  // 圆角
        visible: root.editing
    }

    TextEdit {
       id: textEdit
       text: root.display // 或你的模型数据字段
       color: "black"
       anchors.horizontalCenter: parent.horizontalCenter
       anchors.top: parent.top
       anchors.topMargin: 5
       width: parent.width - 8
       wrapMode: Text.WrapAnywhere
       horizontalAlignment: Text.AlignHCenter
       verticalAlignment: Text.AlignVCenter

       // 只在编辑状态显示
       visible: root.editing
       focus: root.editing

       // 按回车完成编辑
       Keys.onReturnPressed: {
           updateFileName()
       }

       MouseArea {
           anchors.fill: parent
           propagateComposedEvents: true  // 允许事件继续传递
           onClicked: (mouse) => {
               if (parent.text.length > 0) {
                   var cursorPos = textEdit.positionAt(mouse.x, mouse.y);
                   parent.cursorPosition = cursorPos
               }
           }
       }

       onVisibleChanged: {
           if (visible) {
               selectText()
           }
       }

       // 点击外部完成编辑
       onActiveFocusChanged: {
           if (!activeFocus && root.editing) {
               updateFileName()
           }
       }

       // 暴露焦点控制方法
       function forceActiveFocusEx() {
           forceActiveFocus()
       }

       function selectText() {
           var chars = Array.from(text); // 正确处理多字节字符
           var dotIndices = [];

           var isDesktopFile = text.endsWith(".desktop");
           if (!isDesktopFile && !isSoftLink && !isDir && text.includes(".") && !text.startsWith(".")) {
               // 收集所有点号位置（从后向前）
               for (var i = chars.length - 1; i >= 0; i--) {
                   if (chars[i] === '.') {
                       dotIndices.push(i);
                       if (dotIndices.length >= 2) break; // 只需要最后两个点号
                   }
               }
           }
           // 判断是否为双重扩展名（如.tar.gz）
           var isDoubleExtension = false;
           if (dotIndices.length >= 2) {
               var middleExt = text.substring(dotIndices[1], dotIndices[0]);
               isDoubleExtension = middleExt === ".tar" || middleExt === ".zip";
           }

           // 计算选中结束位置
           var selectEnd = dotIndices.length > 0
               ? (isDoubleExtension && dotIndices.length >= 2
                   ? dotIndices[1]  // 对于.tar.gz选中到第一个点号前
                   : dotIndices[0]) // 普通扩展名选中到最后一个点号前
               : text.length;   // 无扩展名选中全部

           // 转换为基于代码单元的位置（兼容QML的select方法）
           var prefix = chars.slice(0, selectEnd).join('');
           var codeUnitEnd = prefix.length;

           // 执行选中并定位光标
           select(0, codeUnitEnd);
       }
   }

   // 外部可通过此方法触发焦点
   function forceActiveFocus() {
       console.log("edit ", root.display)
       textEdit.text = root.display
       root.editing = true
       textEdit.forceActiveFocusEx()
   }

   // 外部可通过此方法触发焦点
   function updateFileName() {
       root.editing = false
       let newName = textEdit.text.trim()

       if (newName === "." || newName === ".." || newName === "") {
           textEdit.text = root.display
           return
       }

       if (newName !== root.display) {
           editFinish(selectedUris, newName)
       }
       console.log("new file name", newName)
   }
}
